"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import pathlib
from datetime import datetime
from glob import glob
from typing import List, Dict, Tuple

import ee
import numpy as np
import pytest
import rasterio as rio
from click.testing import CliRunner
from geedim.cli import cli
from geedim.utils import root_path, asset_id
from rasterio.coords import BoundingBox
from rasterio.crs import CRS
from rasterio.features import bounds
from rasterio.warp import transform_geom
from rasterio.transform import Affine


@pytest.fixture()
def l4_5_image_id_list(l4_image_id, l5_image_id) -> List[str]:
    """ A list of landsat 4 & 5 image ID's. """
    return [l4_image_id, l5_image_id]


@pytest.fixture()
def l8_9_image_id_list(l8_image_id, l9_image_id) -> List[str]:
    """ A list of landsat 8 & 9 image ID's. """
    return [l8_image_id, l9_image_id]


@pytest.fixture()
def s2_sr_image_id_list() -> List[str]:
    """ A list of Sentinel-2 SR image IDs. """
    return [
        'COPERNICUS/S2_SR/20211004T080801_20211004T083709_T34HEJ',
        'COPERNICUS/S2_SR/20211123T081241_20211123T083704_T34HEJ',
        'COPERNICUS/S2_SR/20220107T081229_20220107T083059_T34HEJ'
    ]


@pytest.fixture()
def gedi_image_id_list() -> List[str]:
    """ A list of GEDI canopy top height ID's. """
    return [
        'LARSE/GEDI/GEDI02_A_002_MONTHLY/202008_018E_036S', 'LARSE/GEDI/GEDI02_A_002_MONTHLY/202009_018E_036S',
        'LARSE/GEDI/GEDI02_A_002_MONTHLY/202005_018E_036S'
    ]


def _test_downloaded_file(
    filename: pathlib.Path, region: Dict = None, crs: str = None, scale: float = None, dtype: str = None,
    bands: List[str] = None, scale_offset: bool = None, transform: Affine = None, shape: Tuple[int, int] = None
):
    """ Helper function to test image file format against given parameters. """
    with rio.open(filename, 'r') as ds:
        ds: rio.DatasetReader = ds
        assert ds.nodata is not None
        array = ds.read(masked=True)
        am = array.mean()
        assert np.isfinite(am) and (am != 0)
        if region:
            exp_region = transform_geom('EPSG:4326', ds.crs, region)
            exp_bounds = BoundingBox(*bounds(exp_region))
            assert (
                (ds.bounds[0] <= exp_bounds[0]) and (ds.bounds[1] <= exp_bounds[1]) and
                (ds.bounds[2] >= exp_bounds[2]) and (ds.bounds[3] >= exp_bounds[3])
            )
        if crs:
            assert CRS(ds.crs) == CRS.from_string(crs)
        if scale:
            assert abs(ds.transform[0]) == scale
        if dtype:
            assert ds.dtypes[0] == dtype
        if scale_offset:
            refl_bands = [
                i for i in range(1, ds.count + 1)
                if ('center_wavelength' in ds.tags(i)) and (float(ds.tags(i)['center_wavelength']) < 1)
            ]
            array = ds.read(refl_bands, masked=True)
            assert all(array.min(axis=(1, 2)) >= -0.5)
            assert all(array.max(axis=(1, 2)) <= 1.5)
        if transform:
            assert ds.transform[:6] == transform[:6]
        if shape:
            assert ds.shape == tuple(shape)
        if bands:
            assert set(bands) == set(ds.descriptions)
            assert set(bands) ==  set([ds.tags(bi)['name'] for bi in range(1, ds.count + 1)])


@pytest.mark.parametrize(
    'name, start_date, end_date, region, fill_portion, cloudless_portion, is_csmask', [
        ('LANDSAT/LC09/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha_file', 10, 50, True),
        ('LANDSAT/LE07/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha_file', 0, 0, True),
        ('LANDSAT/LT05/C02/T1_L2', '2005-01-01', '2006-02-01', 'region_100ha_file', 40, 50, True),
        ('COPERNICUS/S2_SR', '2022-01-01', '2022-01-15', 'region_100ha_file', 0, 50, True),
        ('COPERNICUS/S2', '2022-01-01', '2022-01-15', 'region_100ha_file', 50, 40, True),
        ('COPERNICUS/S2_SR_HARMONIZED', '2022-01-01', '2022-01-15', 'region_100ha_file', 0, 50, True),
        ('COPERNICUS/S2_HARMONIZED', '2022-01-01', '2022-01-15', 'region_100ha_file', 50, 40, True),
        ('LARSE/GEDI/GEDI02_A_002_MONTHLY', '2021-11-01', '2022-01-01', 'region_100ha_file', 1, 0, False)
    ]
)
def test_search(
    name, start_date: str, end_date: str, region: str, fill_portion: float, cloudless_portion: float, is_csmask: bool,
    tmp_path: pathlib.Path, runner: CliRunner, request: pytest.FixtureRequest
):
    """
    Test search command gives valid results for different cloud/shadow maskable, and generic collections.
    """
    region_file: Dict = request.getfixturevalue(region)
    results_file = tmp_path.joinpath('search_results.json')
    cli_str = (
        f'search -c {name} -s {start_date} -e {end_date} -r {region_file} -fp {fill_portion} '
        f'-cp {cloudless_portion} -op {results_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (results_file.exists())
    with open(results_file, 'r') as f:
        properties = json.load(f)

    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')
    im_dates = np.array(
        [datetime.utcfromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    # test FILL_PORTION in expected range
    im_fill_portions = np.array([im_props['FILL_PORTION'] for im_props in properties.values()])
    assert np.all(im_fill_portions >= fill_portion) and np.all(im_fill_portions <= 100)
    if is_csmask:  # is a cloud/shadow masked collection
        # test CLOUDLESS_PORTION in expected range
        im_cl_portions = np.array([im_props['CLOUDLESS_PORTION'] for im_props in properties.values()])
        assert np.all(im_cl_portions >= cloudless_portion) and np.all(im_cl_portions <= 100)
    # test search result image dates lie between `start_date` and `end_date`
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)
    # test search result image dates are sorted
    assert np.all(sorted(im_dates) == im_dates)


def test_config_search_s2(region_10000ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test `config` sub-command chained with `search` of Sentinel-2 affects CLOUDLESS_PORTION as expected. """
    results_file = tmp_path.joinpath('search_results.json')
    name = 'COPERNICUS/S2_SR'
    cl_portion_list = []
    for prob in [40, 80]:
        cli_str = (
            f'config --prob {prob} search -c {name} -s 2022-01-01 -e 2022-02-01 -r {region_10000ha_file} -fp -op '
            f'{results_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        assert (results_file.exists())
        with open(results_file, 'r') as f:
            properties = json.load(f)
        cl_portion_list.append(np.array([prop_dict['CLOUDLESS_PORTION'] for prop_dict in properties.values()]))

    assert np.any(cl_portion_list[0] < cl_portion_list[1])
    assert not np.any(cl_portion_list[0] > cl_portion_list[1])


def test_config_search_l9(region_10000ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test `config` sub-command chained with `search` of Landsat-9 affects CLOUDLESS_PORTION as expected. """
    results_file = tmp_path.joinpath('search_results.json')
    name = 'LANDSAT/LC09/C02/T1_L2'
    cl_portion_list = []
    for param in ['--mask-shadows', '--no-mask-shadows']:
        cli_str = (
            f'config {param} search -c {name} -s 2022-02-15 -e 2022-04-01 -r {region_10000ha_file} -fp -op'
            f' {results_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        assert (results_file.exists())
        with open(results_file, 'r') as f:
            properties = json.load(f)
        cl_portion_list.append(np.array([prop_dict['CLOUDLESS_PORTION'] for prop_dict in properties.values()]))

    assert np.any(cl_portion_list[0] < cl_portion_list[1])
    assert not np.any(cl_portion_list[0] > cl_portion_list[1])


def test_region_bbox_search(region_100ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test --bbox gives same search results as --region <geojson file>. """

    results_file = tmp_path.joinpath('search_results.json')
    with open(region_100ha_file, 'r') as f:
        region = json.load(f)
    bbox = bounds(region)
    bbox_str = ' '.join([str(b) for b in bbox])
    cli_strs = [
        f'search -c LANDSAT/LC09/C02/T1_L2 -s 2022-01-01 -e 2022-02-01 -r {region_100ha_file} -op {results_file}',
        f'search -c LANDSAT/LC09/C02/T1_L2 -s 2022-01-01 -e 2022-02-01 -b {bbox_str} -op {results_file}'
    ]

    props_list = []
    for cli_str in cli_strs:
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        assert (results_file.exists())
        with open(results_file, 'r') as f:
            properties = json.load(f)
        props_list.append(properties)

    assert props_list[0] == props_list[1]


def test_raster_region_search(const_image_25ha_file, region_25ha_file, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test --region works with a raster file. """

    results_file = tmp_path.joinpath('search_results.json')
    cli_strs = [
        f'search -c LANDSAT/LC09/C02/T1_L2 -s 2022-01-01 -e 2022-02-01 -r {region_25ha_file} -op {results_file}',
        f'search -c LANDSAT/LC09/C02/T1_L2 -s 2022-01-01 -e 2022-02-01 -r {const_image_25ha_file} -op {results_file}'
    ]

    props_list = []
    for cli_str in cli_strs:
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        assert (results_file.exists())
        with open(results_file, 'r') as f:
            properties = json.load(f)
        props_list.append(properties)

    assert props_list[0].keys() == props_list[1].keys()


def test_search_add_props_l9(region_25ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test --add-property generates results with the additional property keys. """
    results_file = tmp_path.joinpath('search_results.json')
    name = 'LANDSAT/LC09/C02/T1_L2'
    add_props = ['CLOUD_COVER', 'GEOMETRIC_RMSE_VERIFY']
    add_props_str = ''.join([f' -ap {add_prop} ' for add_prop in add_props])
    cli_str = (
        f'search -c {name} -s 2022-01-01 -e 2022-02-01 -r {region_25ha_file} {add_props_str} -op {results_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (results_file.exists())
    with open(results_file, 'r') as f:
        properties = json.load(f)
    prop_keys = list(properties.values())[0].keys()
    assert all([add_prop in prop_keys for add_prop in add_props])
    assert all([add_prop in result.output for add_prop in add_props])


def test_search_custom_filter_l9(region_25ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test --custom-filter filters the search results as specified. """
    results_file = tmp_path.joinpath('search_results.json')
    name = 'LANDSAT/LC09/C02/T1_L2'
    cc_thresh = 30
    add_prop = 'CLOUD_COVER'
    cli_str = (
        f'search -c {name} -s 2022-01-01 -e 2022-04-01 -r {region_25ha_file} -ap {add_prop} '
        f'-cf {add_prop}>{cc_thresh} -op {results_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (results_file.exists())
    with open(results_file, 'r') as f:
        properties = json.load(f)
    prop_keys = list(properties.values())[0].keys()
    assert add_prop in prop_keys
    assert all([prop[add_prop] > cc_thresh for prop in properties.values()])


def test_search_cloudless_portion_no_value(
    region_25ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path
):
    """ Test `search --cloudless-portion` gives the same results as `search --cloudless-portion 0`. """
    results_file = tmp_path.joinpath(f'search_results.json')
    name = 'LANDSAT/LC09/C02/T1_L2'
    clp_list = []
    for post_fix, cp_spec in zip(['no_val', 'zero_val'], ['-cp', '-cp 0']):
        cli_str = (
            f'search -c {name} -s 2022-01-01 -e 2022-04-01 -r {region_25ha_file} {cp_spec} -op {results_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        assert (results_file.exists())
        with open(results_file, 'r') as f:
            props = json.load(f)
        prop_keys = list(props.values())[0].keys()
        assert 'FILL_PORTION' in prop_keys
        assert 'CLOUDLESS_PORTION' in prop_keys
        clp = np.array([prop['CLOUDLESS_PORTION'] for prop in props.values()])
        clp_list.append(clp)
    assert np.all(clp_list[0] == clp_list[1])


@pytest.mark.parametrize(
    'image_id, region_file', [
        ('l8_image_id', 'region_25ha_file'),
        ('s2_sr_hm_image_id', 'region_25ha_file'),
        ('gedi_cth_image_id', 'region_25ha_file'),
        ('modis_nbar_image_id', 'region_25ha_file'),
    ]
)  # yapf: disable
def test_download_region_defaults(
    image_id: str, region_file: pathlib.Path, tmp_path: pathlib.Path, runner: CliRunner, request
):
    """ Test image download with default crs, scale, dtype etc.  """
    image_id = request.getfixturevalue(image_id)
    region_file = request.getfixturevalue(region_file)
    out_file = tmp_path.joinpath(image_id.replace('/', '-') + '.tif')

    cli_str = f'download -i {image_id} -r {region_file} -dd {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (out_file.exists())

    # test downloaded file readability and format
    with open(region_file) as f:
        region = json.load(f)
    _test_downloaded_file(out_file, region=region)


@pytest.mark.parametrize(
    'image_id, region_file', [
        ('l8_image_id', 'region_25ha_file'),
        ('s2_sr_hm_image_id', 'region_25ha_file'),
        ('gedi_cth_image_id', 'region_25ha_file'),
    ]
)  # yapf: disable
def test_download_crs_transform(
    image_id: str, region_file: pathlib.Path, tmp_path: pathlib.Path, runner: CliRunner, request
):
    """ Test image download with crs, crs_transform, & shape specified. """
    image_id = request.getfixturevalue(image_id)
    region_file = request.getfixturevalue(region_file)
    out_file = tmp_path.joinpath(image_id.replace('/', '-') + '.tif')

    # find a transform and shape for region_file
    with open(region_file) as f:
        region = json.load(f)
    region_bounds = bounds(region)
    crs = 'EPSG:4326'
    shape = (11, 12)
    shape_str = ' '.join(map(str, shape))
    crs_transform = rio.transform.from_bounds(*region_bounds, *shape[::-1])
    crs_transform_str = ' '.join(map(str, crs_transform[:6]))

    # run the download
    cli_str = f'download -i {image_id} -c {crs} -ct {crs_transform_str} -sh {shape_str} -dd {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (out_file.exists())

    # test downloaded file readability and format
    _test_downloaded_file(out_file, crs=crs, region=region, transform=crs_transform, shape=shape)


def test_download_like(
    l8_image_id: str, s2_sr_image_id: str, region_25ha_file: pathlib.Path, tmp_path: pathlib.Path, runner: CliRunner
):
    """ Test image download using --like. """
    l8_file = tmp_path.joinpath(l8_image_id.replace('/', '-') + '.tif')
    s2_file = tmp_path.joinpath(s2_sr_image_id.replace('/', '-') + '.tif')

    # download the landsat 8 image to be the template
    l8_cli_str = f'download -i {l8_image_id} -r {region_25ha_file} -dd {tmp_path}'
    result = runner.invoke(cli, l8_cli_str.split())
    assert (result.exit_code == 0)
    assert (l8_file.exists())

    # download the sentinel 2 image like the landsat 8 image
    s2_cli_str = f'download -i {s2_sr_image_id} --like {l8_file} -dd {tmp_path}'
    result = runner.invoke(cli, s2_cli_str.split())
    assert (result.exit_code == 0)
    assert (s2_file.exists())

    # test the landsat 8 image is 'like' the sentinel 2 image
    with rio.open(l8_file) as l8_im, rio.open(s2_file, 'r') as s2_im:
        assert l8_im.crs == s2_im.crs
        assert l8_im.shape == s2_im.shape
        assert l8_im.transform[:6] == s2_im.transform[:6]


@pytest.mark.parametrize(
    'image_id, region_file, crs, scale, dtype, bands, mask, resampling, scale_offset, max_tile_size, max_tile_dim', [
        ('l5_image_id', 'region_25ha_file', 'EPSG:3857', 30, 'uint16', None, False, 'near', False, 16, 10000),
        ('l9_image_id', 'region_25ha_file', 'EPSG:3857', 30, 'float32', None, False, 'near', True, 32, 10000),
        (
            's2_toa_image_id', 'region_25ha_file', 'EPSG:3857', 10, 'float64', ['B5', 'B9'], True, 'bilinear', True, 32,
            10000
        ),
        ('modis_nbar_image_id', 'region_100ha_file', 'EPSG:3857', 500, 'int32', None, False, 'bicubic', False, 4, 100),
        (
            'gedi_cth_image_id', 'region_25ha_file', 'EPSG:3857', 10, 'float32', ['rh99'], True, 'bilinear', False, 32,
            10000
        ),
        ('landsat_ndvi_image_id', 'region_25ha_file', 'EPSG:3857', 30, 'float64', None, True, 'near', False, 32, 10000),
    ]
) # yapf: disable
def test_download_params(
    image_id: str, region_file: str, crs: str, scale: float, dtype: str, bands: List[str], mask: bool, resampling: str,
    scale_offset: bool, max_tile_size: float, max_tile_dim: int, tmp_path: pathlib.Path, runner: CliRunner,
    request: pytest.FixtureRequest
):
    """ Test image download, specifying all cli params except crs_transform and shape. """
    image_id = request.getfixturevalue(image_id)
    region_file = request.getfixturevalue(region_file)
    out_file = tmp_path.joinpath(image_id.replace('/', '-') + '.tif')

    cli_str = (
        f'download -i {image_id} -r {region_file} -dd {tmp_path} --crs {crs} --scale {scale} --dtype {dtype} '
        f'--resampling {resampling} -mts {max_tile_size} -mtd {max_tile_dim}'
    )
    cli_str += ' --mask' if mask else ' --no-mask'
    cli_str += ' --scale-offset' if scale_offset else ' --no-scale-offset'
    cli_str += ''.join([f' --band-name {bn}' for bn in bands]) if bands else ''
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (out_file.exists())

    with open(region_file) as f:
        region = json.load(f)
    # test downloaded file readability and format
    _test_downloaded_file(
        out_file, region=region, crs=crs, scale=scale, dtype=dtype, bands=bands, scale_offset=scale_offset
    )


def test_max_tile_size_error(
    s2_sr_image_id: str, region_100ha_file: pathlib.Path, tmp_path: pathlib.Path, runner: CliRunner, request
):
    """ Test image download with max_tile_size > EE limit raises an EE error.  """
    cli_str = f'download -i {s2_sr_image_id} -r {region_100ha_file} -dd {tmp_path} -mts 100'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert isinstance(result.exception, ValueError)
    assert 'download size limit' in str(result.exception)


def test_max_tile_dim_error(
    s2_sr_image_id: str, region_100ha_file: pathlib.Path, tmp_path: pathlib.Path, runner: CliRunner, request
):
    """ Test image download with max_tile_dim > EE limit raises an EE error.  """
    cli_str = f'download -i {s2_sr_image_id} -r {region_100ha_file} -dd {tmp_path} -mtd 100000'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert isinstance(result.exception, ValueError)
    assert 'download limit' in str(result.exception)


def test_export_drive_params(l8_image_id: str, region_25ha_file: pathlib.Path, runner: CliRunner):
    """ Test export to google drive starts ok, specifying all cli params"""
    cli_str = (
        f'export -i {l8_image_id} -r {region_25ha_file} -df geedim/test --crs EPSG:3857 --scale 30 --dtype uint16 '
        f'--mask --resampling bilinear --no-wait --band-name SR_B4 --band-name SR_B3 --band-name SR_B2'
    )
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)


def test_export_asset_params(l8_image_id: str, region_25ha_file: pathlib.Path, runner: CliRunner):
    """ Test export to asset starts ok, specifying all cli params"""
    # Note when e.g. github runs this test in parallel, it could run into problems trying to overwrite an existing
    # asset.  The overwrite error won't be raised with --no-wait though.  So this test serves at least to check the
    # CLI export options work, and won't fail if run in parallel, even if it runs into overwrite problems.
    folder = f'geedim'
    test_asset_id = asset_id(l8_image_id, folder)
    try:
        ee.data.deleteAsset(test_asset_id)
    except ee.ee_exception.EEException:
        pass

    cli_str = (
        f'export -i {l8_image_id} -r {region_25ha_file} -f {folder} --crs EPSG:3857 --scale 30 --dtype uint16 '
        f'--mask --resampling bilinear --no-wait --type asset --band-name SR_B4 --band-name SR_B3 --band-name SR_B2'
    )
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)


def test_export_asset_no_folder_error(l8_image_id: str, region_25ha_file: pathlib.Path, runner: CliRunner):
    """ Test export to asset raises an error when no folder is specified. """
    cli_str = (
        f'export -i {l8_image_id} -r {region_25ha_file} --crs EPSG:3857 --scale 30 '
        f'--dtype uint16 --mask --resampling bilinear --no-wait --type asset'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0
    assert '--folder' in result.output


@pytest.mark.parametrize('image_list, scale', [('s2_sr_image_id_list', 10), ('l8_9_image_id_list', 30)])
def test_composite_defaults(
    image_list: str, scale: float, region_25ha_file: pathlib.Path, runner: CliRunner, tmp_path: pathlib.Path,
    request: pytest.FixtureRequest
):
    """ Test composite with default CLI parameters.  """
    image_list = request.getfixturevalue(image_list)
    image_ids_str = ' -i '.join(image_list)
    cli_str = f'composite -i {image_ids_str} download --crs EPSG:3857 --scale {scale} -r {region_25ha_file} -dd' \
              f' {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)

    # test downloaded file exists
    out_files = glob(str(tmp_path.joinpath(f'*COMP*.tif')))
    assert len(out_files) == 1

    # test downloaded file readability and format
    with open(region_25ha_file) as f:
        region = json.load(f)
    _test_downloaded_file(out_files[0], region)


@pytest.mark.parametrize(
    'image_list, method, region_file, date, mask, resampling, download_scale', [
        ('s2_sr_image_id_list', 'mosaic', None, '2021-10-01', True, 'near', 10),
        ('l8_9_image_id_list', 'q-mosaic', 'region_25ha_file', None, True, 'bilinear', 30),
        ('l8_9_image_id_list', 'medoid', 'region_25ha_file', None, True, 'near', 30),
        ('gedi_image_id_list', 'medoid', None, None, True, 'bilinear', 25),
    ]
)
def test_composite_params(
    image_list: str, method: str, region_file: str, date: str, mask: bool, resampling: str, download_scale: float,
    region_25ha_file, runner: CliRunner, tmp_path: pathlib.Path, request: pytest.FixtureRequest
):
    """ Test composite with default CLI parameters. """
    image_list = request.getfixturevalue(image_list)
    region_file = request.getfixturevalue(region_file) if region_file else None
    image_ids_str = ' -i '.join(image_list)
    cli_comp_str = f'composite -i {image_ids_str} -cm {method} --resampling {resampling}'
    cli_comp_str += f' -r {region_file}' if region_file else ''
    cli_comp_str += f' -d {date}' if date else ''
    cli_comp_str += ' --mask' if mask else ' --no-mask'
    cli_download_str = f'download -r {region_25ha_file} --crs EPSG:3857 --scale {download_scale} -dd {tmp_path}'
    cli_str = cli_comp_str + ' ' + cli_download_str
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)

    # test downloaded file exists
    out_files = glob(str(tmp_path.joinpath(f'*COMP*.tif')))
    assert len(out_files) == 1

    # test downloaded file readability and format
    with open(region_25ha_file) as f:
        region = json.load(f)
    _test_downloaded_file(out_files[0], region=region, crs='EPSG:3857', scale=download_scale)


def test_search_composite_download(region_25ha_file, runner: CliRunner, tmp_path: pathlib.Path):
    """ Test chaining of `search`, `composite` and `download`. """

    cli_search_str = f'search -c COPERNICUS/S1_GRD -s 2022-01-01 -e 2022-02-01 -r {region_25ha_file}'
    cli_comp_str = f'composite --mask'
    cli_download_str = f'download --crs EPSG:3857 --scale 10 -dd {tmp_path}'
    cli_str = cli_search_str + ' ' + cli_comp_str + ' ' + cli_download_str
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)

    # test downloaded file exists
    out_files = glob(str(tmp_path.joinpath(f'*COMP*.tif')))
    assert len(out_files) == 1

    # test downloaded file readability and format
    with open(region_25ha_file) as f:
        region = json.load(f)
    _test_downloaded_file(out_files[0], region=region, crs='EPSG:3857', scale=10)


def test_search_composite_x2_download(region_25ha_file, runner: CliRunner, tmp_path: pathlib.Path):
    """
    Test chaining of `search`, `composite`, `composite` and `download` i.e. the first composite is included as a
    component image in the second composite.
    """

    cli_search_str = f'search -c l7-c2-l2 -s 2022-01-15 -e 2022-04-01 -r {region_25ha_file} -cp 20'
    cli_comp1_str = f'composite --mask'
    cli_comp2_str = f'composite -i LANDSAT/LE07/C02/T1_L2/LE07_173083_20220103 -cm mosaic --date 2022-04-01 --mask'
    cli_download_str = f'download --crs EPSG:3857 --scale 30 -dd {tmp_path}'
    cli_str = cli_search_str + ' ' + cli_comp1_str + ' ' + cli_comp2_str + ' ' + cli_download_str
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)

    # test downloaded file exists
    out_files = glob(str(tmp_path.joinpath(f'*COMP*.tif')))
    assert len(out_files) == 1

    # test downloaded file readability and format
    with open(region_25ha_file) as f:
        region = json.load(f)
    _test_downloaded_file(out_files[0], region=region, crs='EPSG:3857', scale=30)



"""
   Copyright 2022 Dugal Harris - dugalh@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import ee
from geedim.enums import SpectralDistanceMetric
from typing import Optional, List
"""
    This module contains Medoid related functionality adapted from 'Google Earth Engine tools' under MIT 
    license.  See https://github.com/gee-community/gee_tools.
"""


def sum_distance(
    image: ee.Image, collection: ee.ImageCollection, bands: Optional[List] = None,
    metric: SpectralDistanceMetric = SpectralDistanceMetric.sed,
) -> ee.Image:
    """ Find the sum of the spectral distances between the provided ``image`` and image ``collection``. """
    metric = SpectralDistanceMetric(metric)
    if not bands:
        bands = collection.first().bandNames()
    image = ee.Image(image).select(bands)

    # Notes on masking:
    # - Where ``image`` is masked, the summed distance should be masked.
    # - Where any other image in ``collection`` is masked, the summed distance should omit its contribution.
    #
    # The above requirements are satisfied by leaving images masked, creating an ee.ImageCollection of distances
    # between ``image`` and other images in the collection (these distances are masked where either image is masked),
    # and using ImageCollection.sum() to sum distances, omitting masked areas from the sum.
    # The sum is only masked where all component distances are masked i.e. where ``image`` is masked.

    def accum_dist_to_image(to_image: ee.Image, dist_list: ee.List) -> ee.Image:
        """
        Earth engine iterator function to create a list of spectral distances between ``image`` and ``to_image``.
        """
        to_image = ee.Image(to_image).select(bands)

        # Find the distance between image and to_image.  Both images are not unmasked so that distance will be
        # masked where one or both are masked.
        dist = image.spectralDistance(to_image, metric.value)
        if metric == SpectralDistanceMetric.sed:
            # sqrt scaling is necessary for summing with other distances and equivalence to original method
            dist = dist.sqrt()

        # Append distance to list.
        return ee.List(dist_list).add(dist)

    dist_list = ee.List(collection.iterate(accum_dist_to_image, ee.List([])))
    # TODO: are we better off using mean here to avoid overflow?
    return ee.ImageCollection(dist_list).sum()


def medoid_score(
    collection: ee.ImageCollection, bands: Optional[List] = None, name: str = 'sumdist',
) -> ee.ImageCollection:
    """
    Add medoid score band (i.e. summed distance to all other images) to all images in ``collection``.

    Parameters
    ----------
    collection: ee.ImageCollection
        Collection to add medoid score band to.
    bands: list of str, optional
        Bands to calculate the medoid score from (default: use all bands).
    name: str, optional
        Name of score band to add (default: 'sumdist').

    Returns
    -------
    ee.ImageCollection
        Collection with added medoid score band.
    """

    def add_score_band(image: ee.Image):
        """ Add medoid score band to provided ``image``. """
        image = ee.Image(image)

        # Compute the sum of the euclidean distance between the current image
        # and every image in the rest of the collection
        # TODO: many (~50%) of these distance calcs are duplicates, can we make this more efficient?
        dist = sum_distance(image, collection, bands=bands)

        # multiply by -1 so that highest score is lowest summed distance
        return image.addBands(dist.multiply(-1).rename(name))

    return collection.map(add_score_band)


def medoid(collection: ee.ImageCollection, bands: Optional[List] = None) -> ee.Image:
    """
    Find the medoid composite of an image collection. Adapted from https://www.mdpi.com/2072-4292/5/12/6481, and
    https://github.com/gee-community/gee_tools.

    Parameters
    ----------
    collection: ee.ImageCollection
        Image collection to composite.
    bands: list of str, optional
        Bands to calculate the medoid score from (default: use all bands).

    Returns
    -------
    ee.Image
        Medoid composite image.
    """
    name = 'sumdist'
    medoid_coll = medoid_score(collection, bands, name=name)
    comp_im = medoid_coll.qualityMosaic(name)
    # remove score band and return
    keep_bands = comp_im.bandNames().remove(name)
    return comp_im.select(keep_bands)



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from collections import namedtuple

import ee
import numpy as np
import pytest
from geedim.tile import Tile
from geedim.utils import retry_session
from rasterio import Affine
from rasterio.windows import Window
from tqdm.auto import tqdm

BaseImageLike = namedtuple('BaseImageLike', ['ee_image', 'crs', 'transform', 'shape', 'count', 'dtype'])


@pytest.fixture(scope='module')
def base_image_like(region_25ha):
    """ Create a synthetic image object to emulate BaseImage. """
    ee_image = ee.Image([1, 2, 3]).reproject(crs='EPSG:4326', scale=30).clip(region_25ha)
    ee_info = ee_image.getInfo()
    band_info = ee_info['bands'][0]
    transform = Affine(*band_info['crs_transform']) * Affine.translation(*band_info['origin'])
    return BaseImageLike(ee_image, 'EPSG:3857', transform, tuple(band_info['dimensions'][::-1]), 3, 'uint8')


def test_create(base_image_like):
    """ Test creation of a Tile object that refers to the whole of `base_image_like`. """
    window = Window(0, 0, *base_image_like.shape[::-1])
    tile = Tile(base_image_like, window)
    assert tile.window == window
    assert tile._transform == base_image_like.transform
    assert tile._shape == base_image_like.shape


@pytest.mark.parametrize('session', [None, retry_session()])
def test_download(base_image_like, session):
    """ Test downloading the synthetic image tile.  """
    window = Window(0, 0, *base_image_like.shape[::-1])
    tile = Tile(base_image_like, window)
    dtype_size = np.dtype(tile._exp_image.dtype).itemsize
    raw_download_size = tile._shape[0] * tile._shape[1] * tile._exp_image.count * dtype_size
    bar = tqdm(total=float(raw_download_size))
    array = tile.download(session=session, bar=bar)

    assert array is not None
    assert array.shape == (base_image_like.count, *base_image_like.shape)
    assert array.dtype == np.dtype(base_image_like.dtype)
    for i in range(3):
        assert np.all(array[i] == i + 1)
    assert bar.n == pytest.approx(raw_download_size, rel=0.01)



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from typing import Dict

import ee
import numpy as np
import pytest
import rasterio as rio
from geedim.mask import MaskedImage, get_projection, class_from_id


def test_class_from_id(landsat_image_ids, s2_sr_image_id, s2_toa_hm_image_id, generic_image_ids):
    """  Test class_from_id(). """
    from geedim.mask import LandsatImage, Sentinel2SrClImage, Sentinel2ToaClImage

    assert all([class_from_id(im_id) == LandsatImage for im_id in landsat_image_ids])
    assert all([class_from_id(im_id) == MaskedImage for im_id in generic_image_ids])
    assert class_from_id(s2_sr_image_id) == Sentinel2SrClImage
    assert class_from_id(s2_toa_hm_image_id) == Sentinel2ToaClImage


def test_from_id():
    """ Test MaskedImage.from_id() sets _id. """
    ee_id = 'MODIS/006/MCD43A4/2022_01_01'
    gd_image = MaskedImage.from_id(ee_id)
    assert gd_image._id == ee_id


@pytest.mark.parametrize(
    'masked_image', [
        'user_masked_image', 'modis_nbar_masked_image', 'gch_masked_image', 's1_sar_masked_image',
        'gedi_agb_masked_image', 'gedi_cth_masked_image', 'landsat_ndvi_masked_image'
    ]
)
def test_gen_aux_bands_exist(masked_image: str, request: pytest.FixtureRequest):
    """ Test the presence of auxiliary band (i.e. FILL_MASK) in generic masked images. """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    band_names = masked_image.ee_image.bandNames().getInfo()
    assert 'FILL_MASK' in band_names


@pytest.mark.parametrize(
    'masked_image', [
        's2_sr_masked_image', 's2_toa_masked_image', 's2_sr_hm_masked_image', 's2_toa_hm_masked_image',
        'l9_masked_image', 'l8_masked_image', 'l7_masked_image', 'l5_masked_image', 'l4_masked_image'
    ]
)
def test_cloud_mask_aux_bands_exist(masked_image: str, request: pytest.FixtureRequest):
    """ Test the presence of auxiliary bands in cloud masked images. """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    band_names = masked_image.ee_image.bandNames().getInfo()
    exp_band_names = ['CLOUD_MASK', 'SHADOW_MASK', 'FILL_MASK', 'CLOUDLESS_MASK', 'CLOUD_DIST']
    for exp_band_name in exp_band_names:
        assert exp_band_name in band_names


@pytest.mark.parametrize(
    'masked_image', [
        's2_sr_masked_image', 's2_toa_masked_image', 's2_sr_hm_masked_image', 's2_toa_hm_masked_image',
        'l9_masked_image', 'l8_masked_image', 'l7_masked_image', 'l5_masked_image', 'l4_masked_image',
        'user_masked_image', 'modis_nbar_masked_image', 'gch_masked_image', 's1_sar_masked_image',
        'gedi_agb_masked_image', 'gedi_cth_masked_image', 'landsat_ndvi_masked_image'
    ]
)
def test_set_region_stats(masked_image: str, region_100ha, request: pytest.FixtureRequest):
    """
    Test MaskedImage._set_region_stats() generates the expected properties and that these are in the valid range.
    """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    masked_image._set_region_stats(region_100ha)
    for stat_name in ['FILL_PORTION', 'CLOUDLESS_PORTION']:
        assert stat_name in masked_image.properties
        assert masked_image.properties[stat_name] >= 0 and masked_image.properties[stat_name] <= 100


@pytest.mark.parametrize('image_id', ['l9_image_id', 'l8_image_id', 'l7_image_id', 'l5_image_id', 'l4_image_id'])
def test_landsat_cloudless_portion(image_id: str, request: pytest.FixtureRequest):
    """ Test `geedim` CLOUDLESS_PORTION for the whole image against related Landsat CLOUD_COVER property. """
    image_id: MaskedImage = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_shadows=False, mask_cirrus=False)
    masked_image._set_region_stats()
    # the `geedim` cloudless portion of the filled portion
    cloudless_portion = masked_image.properties['CLOUDLESS_PORTION']
    # landsat provided cloudless portion
    landsat_cloudless_portion = 100 - float(masked_image.properties['CLOUD_COVER'])
    assert cloudless_portion == pytest.approx(landsat_cloudless_portion, abs=5)


@pytest.mark.parametrize('image_id', ['s2_toa_image_id', 's2_sr_image_id', 's2_toa_hm_image_id', 's2_sr_hm_image_id'])
def test_s2_cloudless_portion(image_id: str, request: pytest.FixtureRequest):
    """ Test `geedim` CLOUDLESS_PORTION for the whole image against CLOUDY_PIXEL_PERCENTAGE Sentinel-2 property. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='qa', mask_shadows=False, mask_cirrus=False)
    masked_image._set_region_stats()
    # S2 provided cloudless portion
    s2_cloudless_portion = 100 - float(masked_image.properties['CLOUDY_PIXEL_PERCENTAGE'])
    # CLOUDLESS_MASK is eroded and dilated, so allow 10% difference to account for that
    assert masked_image.properties['CLOUDLESS_PORTION'] == pytest.approx(s2_cloudless_portion, abs=10)


@pytest.mark.parametrize('image_id', ['l9_image_id'])
def test_landsat_cloudmask_params(image_id: str, request: pytest.FixtureRequest):
    """ Test Landsat cloud/shadow masking `mask_shadows` and `mask_cirrus` parameters. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_shadows=False, mask_cirrus=False)
    masked_image._set_region_stats()
    # cloud-free portion
    cloud_only_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True, mask_cirrus=False)
    masked_image._set_region_stats()
    # cloud and shadow-free portion
    cloud_shadow_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True, mask_cirrus=True)
    masked_image._set_region_stats()
    # cloud, cirrus and shadow-free portion
    cloudless_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']

    # test `mask_shadows` and `mask_cirrus` affect CLOUDLESS_PORTION as expected
    assert cloud_only_portion > cloud_shadow_portion
    assert cloud_shadow_portion > cloudless_portion


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_mask_shadows(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `mask_shadows` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_shadows=False)
    masked_image._set_region_stats(region_10000ha)
    # cloud-free portion
    cloud_only_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True)
    masked_image._set_region_stats(region_10000ha)
    # cloud and shadow-free portion
    cloudless_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    assert cloud_only_portion > cloudless_portion


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_prob(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `prob` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True, prob=80)
    masked_image._set_region_stats(region_10000ha)
    prob80_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True, prob=40)
    masked_image._set_region_stats(region_10000ha)
    prob40_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    # test there is more cloud (less CLOUDLESS_PORTION) with prob=40 as compared to prob=80
    assert prob80_portion > prob40_portion


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_method(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `mask_method` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob')
    masked_image._set_region_stats(region_10000ha)
    cloud_prob_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='qa')
    masked_image._set_region_stats(region_10000ha)
    qa_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    # test `mask_method` changes CLOUDLESS_PORTION but not by too much
    assert cloud_prob_portion != qa_portion
    assert cloud_prob_portion == pytest.approx(qa_portion, abs=10)


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_mask_cirrus(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `mask_cirrus` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='qa', mask_cirrus=False)
    # cloud and shadow free portion
    masked_image._set_region_stats(region_10000ha)
    non_cirrus_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='qa', mask_cirrus=True)
    masked_image._set_region_stats(region_10000ha)
    # cloud, cirrus and shadow free portion
    cirrus_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    assert non_cirrus_portion >= cirrus_portion


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_dark(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `dark` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, dark=0.5)
    masked_image._set_region_stats(region_10000ha)
    dark_pt5_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, dark=0.1)
    masked_image._set_region_stats(region_10000ha)
    datk_pt1_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    # test that increasing `dark` results in an increase in detected shadow and corresponding decrease in
    # CLOUDLESS_PORTION
    assert datk_pt1_portion > dark_pt5_portion


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_shadow_dist(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `shadow_dist` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, shadow_dist=200)
    masked_image._set_region_stats(region_10000ha)
    sd200_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, shadow_dist=1000)
    masked_image._set_region_stats(region_10000ha)
    sd1000_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    # test that increasing `shadow_dist` results in an increase in detected shadow and corresponding decrease in
    # CLOUDLESS_PORTION
    assert sd200_portion > sd1000_portion


@pytest.mark.parametrize('image_id', ['s2_sr_image_id', 's2_toa_image_id'])
def test_s2_cloudmask_cdi_thresh(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud/shadow masking `cdi_thresh` parameter. """
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, cdi_thresh=.5)
    masked_image._set_region_stats(region_10000ha)
    cdi_pt5_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    masked_image = MaskedImage.from_id(image_id, cdi_thresh=-.5)
    masked_image._set_region_stats(region_10000ha)
    cdi_negpt5_portion = 100 * masked_image.properties['CLOUDLESS_PORTION'] / masked_image.properties['FILL_PORTION']
    # test that increasing `cdi_thresh` results in an increase in detected cloud and corresponding decrease in
    # CLOUDLESS_PORTION
    assert cdi_negpt5_portion > cdi_pt5_portion


@pytest.mark.parametrize('image_id, max_cloud_dist', [('s2_sr_image_id', 100), ('s2_sr_hm_image_id', 500)])
def test_s2_clouddist_max(image_id: str, max_cloud_dist: int, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test S2 cloud distance `max_cloud_dist` parameter. """

    def get_max_cloud_dist(cloud_dist: ee.Image):
        """ Get the maximum of `cloud_dist` over region_10000ha. """
        mcd = cloud_dist.reduceRegion(reducer='max', geometry=region_10000ha, bestEffort=True, maxPixels=1e4)
        return mcd.get('CLOUD_DIST').getInfo() * 10

    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, max_cloud_dist=max_cloud_dist)
    cloud_dist = masked_image.ee_image.select('CLOUD_DIST')
    meas_max_cloud_dist = get_max_cloud_dist(cloud_dist)
    assert meas_max_cloud_dist == pytest.approx(max_cloud_dist, rel=0.1)


def get_mask_stats(masked_image: MaskedImage, region: Dict):
    """  Get cloud/shadow etc aux band statistics for a given MaskedImage and region. """
    ee_image = masked_image.ee_image
    pan = ee_image.select([0, 1, 2]).reduce(ee.Reducer.mean())
    cdist = ee_image.select('CLOUD_DIST')
    cloud_mask = ee_image.select('CLOUD_MASK')
    shadow_mask = ee_image.select('SHADOW_MASK')
    cloudless_mask = ee_image.select('CLOUDLESS_MASK')
    pan_cloud = pan.updateMask(cloud_mask).rename('PAN_CLOUD')
    pan_shadow = pan.mask(shadow_mask).rename('PAN_SHADOW')
    pan_cloudless = pan.updateMask(cloudless_mask).rename('PAN_CLOUDLESS')
    cdist_cloud = cdist.updateMask(cloud_mask).rename('CDIST_CLOUD')
    cdist_cloudless = cdist.updateMask(cloudless_mask).rename('CDIST_CLOUDLESS')
    stats_image = ee.Image([pan_cloud, pan_shadow, pan_cloudless, cdist_cloud, cdist_cloudless])
    proj = get_projection(ee_image, min_scale=False)
    means = stats_image.reduceRegion(
        reducer='mean', geometry=region, crs=proj.crs(), scale=proj.nominalScale(), bestEffort=True, maxPixels=1e8
    )
    cdist_min = cdist.updateMask(cdist).reduceRegion(
        reducer='min', geometry=region, crs=proj.crs(), scale=proj.nominalScale(), bestEffort=True, maxPixels=1e8
    )
    means = means.set('CDIST_MIN', cdist_min.get('CLOUD_DIST'))
    return means.getInfo()


@pytest.mark.parametrize(
    'masked_image', ['l9_masked_image', 'l8_masked_image', 'l7_masked_image', 'l5_masked_image', 'l4_masked_image']
)
def test_landsat_aux_bands(masked_image: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test Landsat auxiliary band values. """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    stats = get_mask_stats(masked_image, region_10000ha)
    assert stats['PAN_CLOUD'] > stats['PAN_CLOUDLESS']
    assert stats['PAN_CLOUDLESS'] > stats['PAN_SHADOW']
    assert stats['CDIST_CLOUDLESS'] > stats['CDIST_CLOUD']


@pytest.mark.parametrize('masked_image',
    ['s2_sr_masked_image', 's2_toa_masked_image', 's2_sr_hm_masked_image', 's2_toa_hm_masked_image']
)  # yapf: disable
def test_s2_aux_bands(masked_image: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """ Test Sentinel-2 auxiliary band values. """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    stats = get_mask_stats(masked_image, region_10000ha)
    assert stats['PAN_CLOUD'] > stats['PAN_CLOUDLESS']
    assert stats['PAN_CLOUDLESS'] > stats['PAN_SHADOW']
    assert stats['CDIST_CLOUDLESS'] > stats['CDIST_CLOUD']
    proj_scale = get_projection(masked_image.ee_image, min_scale=False).nominalScale().getInfo()
    assert stats['CDIST_MIN'] * 10 == proj_scale


@pytest.mark.parametrize('masked_image', ['s2_sr_masked_image', 'l9_masked_image'])
def test_mask_clouds(masked_image: str, region_100ha: Dict, tmp_path, request: pytest.FixtureRequest):
    """ Test MaskedImage.mask_clouds() by downloading and examining dataset masks. """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    filename = tmp_path.joinpath(f'test_image.tif')
    masked_image.mask_clouds()
    proj_scale = get_projection(masked_image.ee_image, min_scale=False).nominalScale()
    masked_image.download(filename, region=region_100ha, dtype='int32', scale=proj_scale)
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        ds: rio.DatasetReader = ds
        cloudless_mask = ds.read(ds.descriptions.index('CLOUDLESS_MASK') + 1, masked=True)
        assert np.all(cloudless_mask == 1)  # all cloud/shadow areas should be masked (0)
        cloudless_mask = cloudless_mask.filled(0).astype('bool')  # fill nodata with 0 and cast to bool
        # test that cloudless_mask is the same as the nodata/dataset mask for each bands
        ds_masks = ds.read_masks().astype('bool')
        assert np.all(cloudless_mask == ds_masks)


def test_skysat_region_stats():
    """ Test _set_region_stats() works on SKYSAT image with no region. """
    ee_image = ee.Image('SKYSAT/GEN-A/PUBLIC/ORTHO/RGB/s02_20141004T074858Z')
    masked_image = MaskedImage(ee_image)
    masked_image._set_region_stats()
    assert 'FILL_PORTION' in masked_image.properties
    assert masked_image.properties['FILL_PORTION'] > 0.8
##

